import sys,sqlite3
CPI   = 6.0   #Cycles per instruction

class Path:
  def __init__(self,index=0,size=0,speed=0,code='',out=''):
    self.index=index
    self.size=size
    self.speed=speed
    self.code=code
    self.out=out
  def addcode(self,code,size,speed,out):
    self.code+='\n  '+code
    self.size+=size
    self.speed+=speed
    if out!='':
      s=''
      for i in (out+","+self.out.replace("=o5=","=o6=").replace("=o4=","=o5=").replace("=o3=","=o4=").replace("=o2=","=o3=").replace("=o1=","=o2=").replace("=o0=","=o1=")).split(","):
        if i.strip() not in s:
          s+=i+","
      self.out=s
      #print(self.code,s)
      #raw_input()
    return self.score()
  def score(self):
    return (self.size*CPI)**2+self.speed**2
  def __gt__(self,x):
    return self.score()>x.score()
  def __lt__(self,x):
    return self.score()<x.score()
  def __ge__(self,x):
    return self.score()>=x.score()
  def __le__(self,x):
    return self.score()<=x.score()
  def __eq__(self,x):
    return self.score()==x.score()
  def __ne__(self,x):
    return self.score()!=x.score()
  def __str__(self):
    return self.code
  def __call__(self,x):
    return self.index
  def copy(self):
    return Path(self.index, self.size, self.speed, self.code, self.out)
class Node:
  def __init__(self,val='',desc=[]):
    self.value=val
    self.desc=desc
    self.parents=[]
    self.children=[]
  def addparent(self,node):
    node.addhild(self)
    return self
  def addchild(self,node):
    if node not in self.children:
      self.children=[node]+self.children
      if self not in node.parents:
        node.parents+=[self]
    return self
  def child(self,n):
    if n>=len(self.children):
      return Node()
    return self.children[n]
  def parent(self,n):
    return self.parent[n]
  def copy(self):
    n=Node(self.value, self.desc)
    for i in self.children[::-1]:
      n.addchild(i.copy())
    return n
  def __str__(self):
    s=str(self.value)+'['
    for i in self.children:
      s+=str(i)+","
    return (s.rstrip(",")+"]").replace("[]",'')
  def __eq__(self,n):
    if self.desc==n.desc and self.value==n.value:
      f=True
      for i in range(self.desc[0]):
        f&=(self.child(i)==n.child(i))
      return f
    else:
      return False
# start with an empty path
paths=[Path()]

def numargs(s):
  if s in ["+","-","*","/","^","<<","\~"]:
    return 2
  if s in ["++","--"] or s.endswith("("):
    return 1
  return 0
def iscommutative(s):
  return s in ['+','*','min(','max(']
def astgen(l):
  k=numargs(l[-1])
  n=Node(l[-1],desc=[k])
  l=l[0:-1]
  if len(l)==0 or k==0:
    return [n,l]
  while k>0:
    c=astgen(l)
    n.addchild(c[0])
    l=c[1]
    k-=1
  return [n,l]
def isnum(s):
  f=True
  for i in s:
    f&=(("0"<=i<="9")|(i==".")|(i=="-"))
  return f
def astoptimize(n):
  #Optimize each node
  for i in range(n.desc[0]):
    n.children[i]=astoptimize(n.child(i))

  #For commutative operations, put the input with fewest children on the right, or constant inputs
  for i in range(n.desc[0]-1):
    if iscommutative(n.value):
      if n.child(i).desc[0]<n.child(i+1).desc[0] or isnum(n.child(i).value):
        a=n.child(i)
        n.children[i]=n.child(i+1)
        n.children[i+1]=a
  #Perform some replacements. ex, 2*x ==> x<<1
  if n.value=="+" and n.child(0).value=="*" and n.child(1).value=="*":
    if n.child(0).child(0)==n.child(1).child(0):
      #then we have x*y+x*z ==> x*(y+z)
      x=n.child(0).child(0).copy()
      y=n.child(0).child(1).copy()
      z=n.child(1).child(1).copy()
      n=Node("*",[2])
      n.addchild(Node("+",[2]))
      n.children[0].addchild(y)
      n.children[0].addchild(z)
      n.addchild(x)
    elif n.child(0).child(1)==n.child(1).child(1):
      #then we have x*y+x*z ==> x*(y+z)
      x=n.child(0).child(1).copy()
      y=n.child(0).child(0).copy()
      z=n.child(1).child(0).copy()
      n=Node("*",[2])
      n.addchild(Node("+",[2]))
      n.children[0].addchild(y)
      n.children[0].addchild(z)
      n.addchild(x)
    return n
  elif n.value=="++" and "<<" in n.child(0).value:
    n.children[0].value+=" ++"
    return n.child(0)
  elif n.value=="*" and isnum(n.child(1).value):
    if isnum(n.child(0).value):
      return Node(str(int(n.child(0))*int(n.child(1))),[0])
    p=c.execute("SELECT ir FROM z80 WHERE ir=?",(n.child(1).value+" *",)).fetchone()
    p=None
    if p!=None:
      n.value=n.child(1).value+" *"
      n.children=[n.children[0]]
      n.desc=[1]
      return n
    else:
      t=int(n.child(1).value)
      if t==0:
        return Node("0",[0])
      elif t==1:
        return n.child(0)
      elif (t-1)&t==0:
        k=1
        i=2
        while i!=t:
          i+=i
          k+=1
        n.value="<<"
        n.children[1].value=str(k)
        return n
      else:
        return n
  elif n.value=="<<" and isnum(n.child(1).value):
    if int(n.child(1).value)==0:
      return n.child(0)
    p=c.execute("SELECT ir FROM z80 WHERE ir=?",(n.child(1).value+" <<",)).fetchone()
    if p!=None:
      n.value=n.child(1).value+" <<"
      n.children=[n.children[0]]
      n.desc=[1]
    return n
  elif n.value=="+" and n.child(1).value=="0":
    return n.child(0)
  elif n.value=="+" and n.child(1).value=="1":
    n.children=[n.children[0]]
    n.value="++"
    n.desc=[1]
    return n
  elif n.value=="+" and isnum(n.child(1).value):
    if isnum(n.child(0).value):
      return Node(str(int(n.child(0).value)+int(n.child(1).value)),[0])
    p=c.execute("SELECT ir FROM z80 WHERE ir=?",(n.child(1).value+" +",)).fetchone()
    if p!=None:
      n.value=n.child(1).value+" +"
      n.children=[n.children[0]]
      n.desc=[1]
    return n
  else:
    return n
def astoptimize2(n):
  #Optimize each node
  for i in range(n.desc[0]):
    n.children[i]=astoptimize2(n.child(i))

  #For commutative operations, put the input with fewest children on the right, or constant inputs
  for i in range(n.desc[0]-1):
    if iscommutative(n.value):
      if n.child(i).desc[0]<n.child(i+1).desc[0] or isnum(n.child(i).value):
        a=n.child(i)
        n.children[i]=n.child(i+1)
        n.children[i+1]=a
  #Perform some replacements. ex, 2*x ==> x<<1
  if n.value=="*" and isnum(n.child(1).value):
    if isnum(n.child(0).value):
      return Node(str(int(n.child(0))*int(n.child(1))),[0])
    p=c.execute("SELECT ir FROM z80 WHERE ir=?",(n.child(1).value+" *",)).fetchone()
    p=None
    if p!=None:
      n.value=n.child(1).value+" *"
      n.children=[n.children[0]]
      n.desc=[1]
      return n
    else:
      t=int(n.child(1).value)
      if t==0:
        return Node("0",[0])
      elif t==1:
        return n.child(0)
      elif (t-1)&t==0:
        k=1
        i=2
        while i!=t:
          i+=i
          k+=1
        n.value="<<"
        n.children[1].value=str(k)
        return n
      elif (((t-1)&t)-1)&(t-1)&t==0:
        #multiply by 2^m(2^l+1)
        k=t&(t-1)
        t-=k
        m=0
        i=1
        while i!=t:
          m+=1
          i+=i
        l=-m
        i=1
        while i!=k:
          l+=1
          i+=i
        #So I have *(x,2^m(2^l+1)) == > (x+x<<l)<<m
        n0=Node("<<",[2])
        n0.children=[n.child(0).copy(),Node(str(l),[0])]
        n1=Node("+",[2])
        n1.children=[n.child(0).copy(),n0]
        n2=Node("<<",[2])
        n2.children=[n1,Node(str(m),[0])]
        return n2
      else:
        return n
  else:
    return n
def asttorpn(n):
  l=[n.value]
  for i in n.children[::-1]:
    l=asttorpn(i)+l
  return l
def movreg(o,i):
  o=o.strip()
  i=i.strip()
  if o==i:
    return ''
  if o.endswith(i):
    return ''
  if (o.endswith("HL") and i.endswith("DE")) or (o.endswith("DE") and i.endswith("HL")):
    return " \ ex de,hl"
  elif o.endswith("HL") and i.endswith("BC"):
    return " \ ld b,h \ ld c,l"
  elif o.endswith("DE") and i.endswith("BC"):
    return " \ ld b,d \ ld c,e"
  elif o.endswith("BC") and i.endswith("DE"):
      return " \ ld d,b \ ld e,c"
  elif o.endswith("BC") and i.endswith("HL"):
    return " \ ld h,b \ ld l,c"
  else:
    s="Can't move regs!\n           "+o+" ==> "+i
    raise Exception(s)
def astcompile(n,outp=''):
  #We need to compile this node so that it has the right output
  global c

  # Get all of the options for this tree
  p=c.execute("SELECT code,input,output,destroys FROM z80 WHERE ir=? AND output LIKE ? ORDER BY size*size*36+speed*speed ASC",(n.value,"%"+outp+"=%",)).fetchall()
  if p==[]:
    p=c.execute("SELECT code,input,output,destroys FROM z80 WHERE ir=? ORDER BY size*size*36+speed*speed ASC",(n.value,)).fetchall()
    if p==[]:
      if isnum(n.value):
        p=c.execute("SELECT code,input,output,destroys FROM z80 WHERE ir=? AND output LIKE ? ORDER BY size*size*36+speed*speed ASC",("constant","%"+outp+"=%",)).fetchall()
      else:
        p=c.execute("SELECT code,input,output,destroys FROM z80 WHERE ir=? AND output LIKE ? ORDER BY size*size*36+speed*speed ASC",("var","%"+outp+"=%",)).fetchall()
      if p==[]:
        raise Exception(n)
  p=p[0]
  #print(n.value,outp,p)
  #p=c.execute("SELECT code,input,output,destroys FROM z80 WHERE ir=? ORDER BY size*size*36+speed*speed ASC",(n.value,)).fetchone()   #In the future choose fetchall() and try all of the paths.
  n.destroys=[]
  k=0
  ins=p[1].split(",")
  while k<len(n.children):
    n.children[k]=astcompile(n.child(k),ins[k].strip().split("=")[0][-2:])
    for j in n.children[k].destroys:
      if j not in n.destroys:
        n.destroys+=[j]
    k+=1
  if n.desc[0]==0:
    n.code=p[0].replace("**",n.value)
    n.outp=[p[2]]
    if p[2] not in n.destroys:
      s=p[2].split("=")[0].split(":")
      n.destroys+=s
    return n

  for i in p[3]:
    if i in n.destroys:
      n.destroys+=[i]

  #We have the code, now we need to match inputs!
  code=p[0]
  inp=p[1].split(",")
  n.outp=p[2].split(",")

  if inp==[]:
    return n
  
  n.code=n.child(0).code
  if len(inp)==1:
    o=n.child(0).outp[0].split("=")
    i=inp[0].split("=")
    if o[0].endswith(i[0]):
      n.code+=" \ "+code
      return n
    n.code+=movreg(o[0],i[0])+" \ "+code
    return n
  elif len(inp)==2:
    o1=n.child(0).outp[0].split("=")
    i1=inp[0].split("=")
    o2=n.child(1).outp[0].split("=")
    i2=inp[1].split("=")
    if (o1[0][-2:] in n.child(1).destroys) or (o1[0]==o2[0]):
      f=(o1[0][-2:]=="HL" and "DE" not in n.child(1).destroys) or (o1[0][-2:]=="DE" and "HL" not in n.child(1).destroys)
      if f:
        n.code+=" \ ex de,hl"
      else:
        n.code+=" \ push "+o1[0][-2:].lower()
      n.code+=" \ "+n.child(1).code
      m=movreg(o2[0],i2[0])
      if (("ld b," in m) or "ld c," in m) and ("BC" not in n.destroys):
        n.destroys+=["BC"]
      if (("ld d," in m) or "ld e," in m) and ("DE" not in n.destroys):
        n.destroys+=["DE"]
      if (("ld h," in m) or "ld l," in m) and ("HL" not in n.destroys):
        n.destroys+=["HL"]
      if (("ld ixh," in m) or "ld ixl," in m) and ("IX" not in n.destroys):
        n.destroys+=["IX"]

      # Now move the outputs to the right inputs
      n.code+=m
      
      #Now pop and go
      if f:
        n.code+=" \ ex de,hl"
      else:
        n.code+=" \ pop "+i1[0].lower()
      n.code+=" \ "+code
      return n
    else:
      n.code+=" \ "+n.child(1).code
      # Now move the outputs to the right inputs
      m=movreg(o2[0],i2[0])
      if (("ld b," in m) or "ld c," in m) and ("BC" not in n.destroys):
        n.destroys+=["BC"]
      if (("ld d," in m) or "ld e," in m) and ("DE" not in n.destroys):
        n.destroys+=["DE"]
      if (("ld h," in m) or "ld l," in m) and ("HL" not in n.destroys):
        n.destroys+=["HL"]
      if (("ld ixh," in m) or "ld ixl," in m) and ("IX" not in n.destroys):
        n.destroys+=["IX"]
      n.code+=m
      m=movreg(o1[0],i1[0])
      if (("ld b," in m) or "ld c," in m) and ("BC" not in n.destroys):
        n.destroys+=["BC"]
      if (("ld d," in m) or "ld e," in m) and ("DE" not in n.destroys):
        n.destroys+=["DE"]
      if (("ld h," in m) or "ld l," in m) and ("HL" not in n.destroys):
        n.destroys+=["HL"]
      if (("ld ixh," in m) or "ld ixl," in m) and ("IX" not in n.destroys):
        n.destroys+=["IX"]
      n.code+=m+" \ "+code
      return n
  else:
    raise Exception("Not handling more than two arguments!")

def compile():
  global src,c,paths
  mx=0
  out=Path(size=1<<32,speed=1<<64)
  while len(paths)>1 or out.index==0:
    #Grab the next instruction
    i=paths[0].index
    i=src[i:i+3]
    p=c.execute("SELECT code,size,speed,input,output,requires,state FROM z80 WHERE ir=?",(i[0]+" "+i[1]+" "+i[2],)).fetchall()
    #print(str(paths[0]),paths[0].index,i)
    #raw_input()
    if p!=[]:
      #Now we need to find all of the potential paths
      for i in p:
        t=True
        for j in i[3].replace("=i","=o").split(","):
          t&=(j.strip()[0:-4] in paths[0].out)
        if t:
          pn=paths[0].copy()
          pn.addcode(i[0],i[1],i[2],i[4])
          pn.index+=3
          paths+=[pn]
        
      paths=paths[1:]
    else:
      p=c.execute("SELECT code,size,speed,input,output,requires,state FROM z80 WHERE ir=?",(i[0]+" "+i[1],)).fetchall()
      if p!=[]:
        #Now we need to find all of the potential paths
        for i in p:
          t=True
          for j in i[3].replace("=i","=o").split(","):
            t&=(j.strip()[0:-4] in paths[0].out)
          if t:
            pn=paths[0].copy()
            pn.addcode(i[0],i[1],i[2],i[4])
            pn.index+=2
            paths+=[pn]
        paths=paths[1:]
      elif i[1]=="\\~":
        for j in c.execute("SELECT code,size,speed,input,output,requires,state FROM z80 WHERE ir=?",("\\~",)).fetchall():
          j=list(j)
          j[0]=j[0].replace("**","var_"+i[0])
          p+=[j]
          for i in p:
            t=True
            for j in i[3].replace("=i","=o").split(","):
              t&=(j.strip()[0:-4] in paths[0].out)
            if t:
              pn=paths[0].copy()
              pn.addcode(i[0],i[1],i[2],i[4])
              pn.index+=2
              paths+=[pn]
          paths=paths[1:]
      else:
        p=c.execute("SELECT code,size,speed,input,output,requires,state FROM z80 WHERE ir=?",(i[0],)).fetchall()
        if p==[]:
          if "0"<=i[0]<="9":
            p=[]
            for j in c.execute("SELECT code,size,speed,input,output,requires,state FROM z80 WHERE ir=?",("constant",)).fetchall():
              j=list(j)
              j[0]=j[0].replace("**",i[0])
              p+=[j]
          else:
            p=[]
            for j in c.execute("SELECT code,size,speed,input,output,requires,state FROM z80 WHERE ir=?",("var",)).fetchall():
              j=list(j)
              j[0]=j[0].replace("**","var_"+i[0])
              p+=[j]
        #Now we need to find all of the potential paths
        for i in p:
          t=True
          for j in i[3].replace("=i","=o").split(","):
            if j.strip()!='':
              t&=(j.strip()[0:-4] in paths[0].out)
          if t:
            pn=paths[0].copy()
            pn.addcode(i[0],i[1],i[2],i[4])
            pn.index+=1
            paths+=[pn]
        paths=paths[1:]
    paths=sorted(paths)
    if mx<len(paths):
      mx=len(paths)
    if paths==[]:
      paths=[Path()]
    if paths[0].index>=len(src)-2:
      if out>paths[0]:
        out=paths[0]
    while (paths[0]>=out) and len(paths)>1:
      paths=paths[1:]
  print("Searched %d paths" % mx)
  return out
fi=''
fo=''
mode=''
asm=False
if len(sys.argv)==1:
  print("%s source [dest]" % sys.argv[0])
  raise SystemExit
for i in sys.argv[1:]:
  if i[0]=='-':
    pass
  else:
    if fi=='':
      fi=i
    else:
      fo=i
if fo=='':
  s=fi.split('.')
  fo=s[0]
  for i in s[1:-1]:
    fo+='.'+i
  fo+='.asm'
print("Generating Z80 code from %s" % fi)
f=open(fi,"r")
code=f.read()
f.close()

# Split the input
src=code.strip().split('\t')

# Load the description database
conn=sqlite3.connect('irToZ80.db')
c=conn.cursor()

ast=astgen(src)[0]
print(ast)
s=""
while s!=asttorpn(ast):
  s=asttorpn(ast)
  ast=astoptimize(ast)
  ast=astoptimize2(ast)
print(ast)
ast=astcompile(ast)
#cmpld=compile()

#Close the database
conn.close()

f=open(fo,'w')
#f.write(cmpld.code)
f.write((" "+ast.code).replace(" \\",'\n'))
f.close()
print("Output Z80 code to %s" % fo)
#print("%d bytes" % cmpld.size)